# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from functools import cached_property, lru_cache

import contrast
from contrast.agent import agent_state, request_state, scope as scope_
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_middleware import (
    BaseMiddleware,
    log_request_start_and_end,
)
from contrast.agent.asgi import (
    track_scope_sources,
    ASGIRequest,
    ASGIResponse,
)
from contrast.agent.request_context import RequestContext
from contrast.utils.safe_import import safe_import_list
from contrast.utils.assess.duck_utils import safe_getattr_list

from contrast.utils.monitoring import Profiler, tracer
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")

DEFAULT_ASGI_NAME = "asgi_app"


@lru_cache(maxsize=1)
def _get_quart_types():
    return tuple(safe_import_list("quart.Quart"))


class ASGIMiddleware(BaseMiddleware):
    @scope_.contrast_scope()
    def __init__(self, app, app_name=None, original_app=None):
        """
        Contrast middleware - ASGI 3.0

        The `app` argument here must be named "app" exactly to support FastAPI's
        middleware system, which uses the kwarg when calling this method.

        We need to keep the `original_app` kwarg for now to prevent a breaking API
        change.
        """
        del original_app

        # TODO: PYT-2852 Revisit application name detection
        self.app_name = (
            app_name
            if app_name is not None
            else safe_getattr_list(
                app,
                [
                    "__name__",
                    "name",
                    "title",
                ],
                DEFAULT_ASGI_NAME,
            )
        )

        super().__init__()

        if not agent_state.in_automatic_middleware() and isinstance(
            app, _get_quart_types()
        ):
            # we need this to prevent a breaking API change when wrapping Quart apps
            app = app.asgi_app

        self.asgi_app = app

    async def __call__(self, scope, receive, send) -> None:
        if request_state.get_request_id() is not None:
            # This can happen if a single app is wrapped by multiple instances of the
            # middleware (usually caused by automatic instrumentation)
            logger.debug("Detected preexisting request_id - passing through")
            await self.asgi_app(scope, receive, send)
            return

        request_path = scope.get("path", "")

        if scope.get("type") != "http":
            logger.debug("Detected non-http request - cannot analyze", scope=scope)
            await self.call_without_agent_async(scope, receive, send)
            return

        # the request_id context manager must come first!
        with (
            request_state.request_id_context(),
            Profiler(request_path, self.settings.config),
            log_request_start_and_end(
                scope.get("method", ""),
                request_path,
            ),
            tracer(request_path, self.settings.config),
        ):
            request = ASGIRequest(scope, receive)
            environ = await request.to_wsgi_environ()

            context = self.should_analyze_request(environ)
            if context:
                with contrast.CS__CONTEXT_TRACKER.lifespan(context):
                    await self.call_with_agent(context, request, send)
                    return

            await self.call_without_agent_async(scope, request.contrast__receive, send)

    async def call_with_agent(self, context: RequestContext, request, send) -> None:
        with scope_.contrast_scope():
            track_scope_sources(context, request.scope)

            try:
                self.prefilter()

                response = ASGIResponse(send)
                with scope_.pop_contrast_scope():
                    await self.asgi_app(
                        request.scope,
                        request.contrast__receive,
                        response.contrast__send,
                    )
                    context.extract_response(response)

                self.postfilter(context)
                self.check_for_blocked(context)

            finally:
                self.handle_ensure(context, context.request)
                if context.assess_enabled:
                    contrast.STRING_TRACKER.ageoff()

    async def call_without_agent_async(self, scope, receive, send) -> None:
        with scope_.contrast_scope():
            await self.asgi_app(scope, receive, send)

    @cached_property
    def trigger_node(self):
        """
        trigger node used by reflected xss postfilter rule
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.asgi_app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    @cached_property
    def name(self):
        return "asgi"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import argparse
import os
import sys
import shutil

from contrast import __file__
from contrast.configuration.agent_config import AgentConfig
from contrast_rewriter import REWRITE_FOR_PYTEST
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")

DESCRIPTION = """
The command-line runner for the Contrast Python Agent.
"""

USAGE = "%(prog)s [-h] -- cmd [cmd ...]"

EPILOG = """
Insert this command before the one you usually use to start your webserver
to apply Contrast's instrumentation. See our public documentation for details:
https://docs.contrastsecurity.com/en/python.html
"""


def runner() -> None:
    logger.info("Starting Contrast Agent runner pre-process")
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        usage=USAGE,
        epilog=EPILOG,
    )
    parser.add_argument(
        "--rewrite-for-pytest", action="store_true", help=argparse.SUPPRESS
    )
    # if you add public arguments here, update USAGE accordingly
    parser.add_argument(
        "cmd",
        nargs="+",
        help="Command to run with the agent. cmd should be available in the operating system PATH.",
    )

    parsed = parser.parse_args()

    config = AgentConfig()

    loader_path = os.path.join(os.path.dirname(__file__), "loader")
    os.environ["PYTHONPATH"] = os.path.pathsep.join([loader_path] + sys.path)
    os.environ["CONTRAST_INSTALLATION_TOOL"] = "CONTRAST_PYTHON_RUN"

    if parsed.rewrite_for_pytest or config.should_pytest_rewrite:
        os.environ[REWRITE_FOR_PYTEST] = "true"

    cmd_path, *args = shutil.which(parsed.cmd[0]), *parsed.cmd
    if cmd_path is None:
        logger.error("Command not found in PATH", cmd=parsed.cmd[0])
        logger.info(f"Run '{sys.argv[0]} --help' for usage information.")
        sys.exit(1)
        return
    os.execl(cmd_path, *args)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from copy import copy
from typing import TypedDict
from urllib.parse import parse_qs, urlencode, unquote
from collections.abc import Mapping

from contrast.api.attack import ProtectResponse
from contrast_vendor import structlog as logging
from contrast_vendor.webob.multidict import MultiDict

logger = logging.getLogger("contrast")

MASK = "contrast-redacted-{}"
BODY_MASK = b"contrast-redacted-body"
SEMICOLON_URL_ENCODE_VAL = "%25"


class SensitiveDataRule(TypedDict):
    id: str
    keywords: list[str]


class SensitiveDataPolicy(TypedDict):
    mask_attack_vector: bool
    mask_http_body: bool
    rules: list[SensitiveDataRule]


class RequestMasker:
    def __init__(self, config: Mapping):
        self.request = None
        self.attacks = None
        self.mask_rules: SensitiveDataPolicy = config.get(
            "application.sensitive_data_masking_policy", None
        )

    @classmethod
    def new_request_masker(cls, config: SensitiveDataPolicy):
        return cls({"application.sensitive_data_masking_policy": config})

    def mask_sensitive_data(self, request, attacks=None):
        if not request or not self.mask_rules:
            return

        self.request = request
        self.attacks = attacks or []

        logger.debug("Masker: masking sensitive data")

        self._mask_body()
        self._mask_query_string()
        self._mask_request_params()
        self._mask_request_cookies()
        self._mask_request_headers()

        request._masked = True

    def _mask_body(self):
        # Check if mask_http_body is set to False or is None and skip if true
        if not self.mask_rules.get("mask_http_body"):
            return

        # Checks if body is not empty or null
        if self.request.body:
            self.request._masked_body = BODY_MASK

    def _mask_query_string(self):
        if self.request.query_string:
            self.request._masked_query_string = self._mask_raw_query(
                self.request.query_string
            )

    def _mask_raw_query(self, query_string):
        qs_dict = parse_qs(query_string)
        masked_qs_dict = self._mask_dictionary(qs_dict)
        return urlencode(masked_qs_dict, doseq=True)

    def _mask_request_params(self):
        try:
            params = self.request.params
        except ValueError as e:
            # This likely came form a malformed multipart/form-data request,
            # which is out of our control. Still log the error in case
            # there's another case we haven't considered.
            logger.debug("WARNING: failed to parse params", error=e)
            # set the _masked_params to avoid raising the same error
            # in reportable_request.
            self.request._masked_params = {}
            return
        if not params:
            return

        self.request._masked_params = self._mask_dictionary(params)

    def _mask_request_cookies(self):
        cookies = self.request.cookies
        if not cookies:
            return

        self.request._masked_cookies = self._mask_dictionary(cookies)

    def _mask_request_headers(self):
        headers = self.request.headers
        if not headers:
            return

        self.request._masked_headers = self._mask_dictionary(headers)

    def _mask_dictionary(self, d):
        if not d:
            return d

        if isinstance(d, MultiDict):
            d = d.mixed()

        d_copy = {k: copy(v) for k, v in d.items()}

        for k, v in d_copy.items():
            if k is None or self._find_value_index_in_rules(k.lower()) == -1:
                continue

            if isinstance(v, list):
                self._mask_values(k, v, d_copy, self.attacks)
            else:
                self._mask_hash(k, v, d_copy, self.attacks)
        return d_copy

    def _mask_values(self, k, v, d, attacks):
        for idx, item in enumerate(v):
            if self.mask_rules.get("mask_attack_vector") and self._is_value_vector(
                attacks, item
            ):
                d[k][idx] = MASK.format(k.lower())
            if not self._is_value_vector(attacks, item):
                d[k][idx] = MASK.format(k.lower())

    def _mask_hash(self, k, v, d, attacks):
        if self.mask_rules.get("mask_attack_vector") and self._is_value_vector(
            attacks, v
        ):
            d[k] = MASK.format(k.lower())
        if not self._is_value_vector(attacks, v):
            d[k] = MASK.format(k.lower())

    def _is_value_vector(self, attacks, value):
        if not attacks or not value:
            return False

        for attack in attacks:
            if self._is_value_in_sample(attack.samples, value):
                return attack.response != ProtectResponse.NO_ACTION

        return False

    def _is_value_in_sample(self, samples, value):
        if not samples:
            return False

        # Setting this to remove url encoding of header and cookie values
        value = unquote(value)

        for sample in samples:
            if sample.user_input.value == value:
                return True
        return False

    def _find_value_index_in_rules(self, s):
        index = -1
        # When looking for header it replaces '_' with '-' and I don't want to risk not
        # properly matching to the rules
        s = s.replace("-", "_")
        for rule in self.mask_rules.get("rules"):
            try:
                index = rule.get("keywords").index(s)
                break
            except ValueError:
                index = -1

        return index



# coding: utf-8

from _ruamel_yaml import CParser, CEmitter  # type: ignore

from contrast_vendor.ruamel.yaml.constructor import Constructor, BaseConstructor, SafeConstructor
from contrast_vendor.ruamel.yaml.representer import Representer, SafeRepresenter, BaseRepresenter
from contrast_vendor.ruamel.yaml.resolver import Resolver, BaseResolver


from typing import Any, Union, Optional  # NOQA
from contrast_vendor.ruamel.yaml.compat import StreamTextType, StreamType, VersionType  # NOQA

__all__ = ['CBaseLoader', 'CSafeLoader', 'CLoader', 'CBaseDumper', 'CSafeDumper', 'CDumper']


# this includes some hacks to solve the  usage of resolver by lower level
# parts of the parser


class CBaseLoader(CParser, BaseConstructor, BaseResolver):  # type: ignore
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = None,
        preserve_quotes: Optional[bool] = None,
    ) -> None:
        CParser.__init__(self, stream)
        self._parser = self._composer = self
        BaseConstructor.__init__(self, loader=self)
        BaseResolver.__init__(self, loadumper=self)
        # self.descend_resolver = self._resolver.descend_resolver
        # self.ascend_resolver = self._resolver.ascend_resolver
        # self.resolve = self._resolver.resolve


class CSafeLoader(CParser, SafeConstructor, Resolver):  # type: ignore
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = None,
        preserve_quotes: Optional[bool] = None,
    ) -> None:
        CParser.__init__(self, stream)
        self._parser = self._composer = self
        SafeConstructor.__init__(self, loader=self)
        Resolver.__init__(self, loadumper=self)
        # self.descend_resolver = self._resolver.descend_resolver
        # self.ascend_resolver = self._resolver.ascend_resolver
        # self.resolve = self._resolver.resolve


class CLoader(CParser, Constructor, Resolver):  # type: ignore
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = None,
        preserve_quotes: Optional[bool] = None,
    ) -> None:
        CParser.__init__(self, stream)
        self._parser = self._composer = self
        Constructor.__init__(self, loader=self)
        Resolver.__init__(self, loadumper=self)
        # self.descend_resolver = self._resolver.descend_resolver
        # self.ascend_resolver = self._resolver.ascend_resolver
        # self.resolve = self._resolver.resolve


class CBaseDumper(CEmitter, BaseRepresenter, BaseResolver):  # type: ignore
    def __init__(
        self: StreamType,
        stream: Any,
        default_style: Any = None,
        default_flow_style: Any = None,
        canonical: Optional[bool] = None,
        indent: Optional[int] = None,
        width: Optional[int] = None,
        allow_unicode: Optional[bool] = None,
        line_break: Any = None,
        encoding: Any = None,
        explicit_start: Optional[bool] = None,
        explicit_end: Optional[bool] = None,
        version: Any = None,
        tags: Any = None,
        block_seq_indent: Any = None,
        top_level_colon_align: Any = None,
        prefix_colon: Any = None,
    ) -> None:
        # NOQA
        CEmitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            encoding=encoding,
            allow_unicode=allow_unicode,
            line_break=line_break,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags,
        )
        self._emitter = self._serializer = self._representer = self
        BaseRepresenter.__init__(
            self,
            default_style=default_style,
            default_flow_style=default_flow_style,
            dumper=self,
        )
        BaseResolver.__init__(self, loadumper=self)


class CSafeDumper(CEmitter, SafeRepresenter, Resolver):  # type: ignore
    def __init__(
        self: StreamType,
        stream: Any,
        default_style: Any = None,
        default_flow_style: Any = None,
        canonical: Optional[bool] = None,
        indent: Optional[int] = None,
        width: Optional[int] = None,
        allow_unicode: Optional[bool] = None,
        line_break: Any = None,
        encoding: Any = None,
        explicit_start: Optional[bool] = None,
        explicit_end: Optional[bool] = None,
        version: Any = None,
        tags: Any = None,
        block_seq_indent: Any = None,
        top_level_colon_align: Any = None,
        prefix_colon: Any = None,
    ) -> None:
        # NOQA
        self._emitter = self._serializer = self._representer = self
        CEmitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            encoding=encoding,
            allow_unicode=allow_unicode,
            line_break=line_break,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags,
        )
        self._emitter = self._serializer = self._representer = self
        SafeRepresenter.__init__(
            self, default_style=default_style, default_flow_style=default_flow_style
        )
        Resolver.__init__(self)


class CDumper(CEmitter, Representer, Resolver):  # type: ignore
    def __init__(
        self: StreamType,
        stream: Any,
        default_style: Any = None,
        default_flow_style: Any = None,
        canonical: Optional[bool] = None,
        indent: Optional[int] = None,
        width: Optional[int] = None,
        allow_unicode: Optional[bool] = None,
        line_break: Any = None,
        encoding: Any = None,
        explicit_start: Optional[bool] = None,
        explicit_end: Optional[bool] = None,
        version: Any = None,
        tags: Any = None,
        block_seq_indent: Any = None,
        top_level_colon_align: Any = None,
        prefix_colon: Any = None,
    ) -> None:
        # NOQA
        CEmitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            encoding=encoding,
            allow_unicode=allow_unicode,
            line_break=line_break,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags,
        )
        self._emitter = self._serializer = self._representer = self
        Representer.__init__(
            self, default_style=default_style, default_flow_style=default_flow_style
        )
        Resolver.__init__(self)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Using logger.info/debug/someOtherLevel() is not supported in this module. In order to get the correct
frame info, we must skip over functions called in this module and in vendored structlog. If logging is attempted,
incorrect frame info will be displayed on the log message if used in this file.

Use print(...) instead
"""

import asyncio
import os
import pathlib
import io
import sys
from functools import partial
from typing import Union, TextIO, Optional

from contrast.agent import request_state
from contrast.utils.namespace import Namespace
from contrast_vendor import structlog
from contrast.utils.configuration_utils import get_hostname
from . import DEFAULT_PROGNAME, DEFAULT_AGENT_LOGGER_PATH


class module(Namespace):
    file_handle: Optional[io.IOBase] = DEFAULT_AGENT_LOGGER_PATH


def add_hostname(logger, method_name, event_dict):
    event_dict["hostname"] = get_hostname()
    return event_dict


def add_request_id(logger, method_name, event_dict):
    event_dict["request_id"] = request_state.get_request_id()
    return event_dict


def rename_key(old_name, new_name):
    def key_renamer(logger, method_name, event_dict):
        value = event_dict.get(old_name)
        if value and not event_dict.get(new_name):
            event_dict[new_name] = value
            del event_dict[old_name]

        return event_dict

    return key_renamer


def add_progname(logger, method_name, event_dict, progname=DEFAULT_PROGNAME):
    """
    progname is the name of the process the agents uses in logs.
    The default value is Contrast Agent. progname will be used
    as the name of the logger as seen in the logs.
    """
    event_dict["name"] = progname

    return event_dict


def add_asyncio_info(logger, method_name, event_dict):
    try:
        current_task = asyncio.current_task()

        # If no name has been explicitly assigned to the Task, the default asyncio Task implementation
        # generates a default name during instantiation.
        event_dict["asyncio_task_name"] = current_task.get_name()

        current_coro = current_task.get_coro()
        if hasattr(current_coro, "__name__"):
            event_dict["asyncio_coro_name"] = current_coro.__name__

        event_dict["asyncio_task_id"] = id(current_task)
    except Exception:
        # This can happen when there is no running event loop
        pass

    return event_dict


def _log_level_to_int(level: str) -> int:
    level_to_int = {
        "DEBUG": 10,
        "TRACE": 10,
        "INFO": 20,
        "WARNING": 30,
        "ERROR": 40,
        "CRITICAL": 50,
    }

    level = level.upper()
    if level == "TRACE":
        sys.stderr.write(
            "Contrast Python Agent: TRACE logging is equivalent to DEBUG\n"
        )

    return level_to_int.get(level, level_to_int["INFO"])


def _close_handler():
    ignore_fds = (sys.stderr, sys.stdout, None)
    file_name = getattr(module.file_handle, "name", None)
    # ignore_names handles a specific case for capturing sys.stderr/stdout in pytest
    # see also: https://docs.pytest.org/en/stable/how-to/capture-stdout-stderr.html
    ignore_names = ("<stderr>", "<stdout>")

    if (
        module.file_handle not in ignore_fds
        and file_name not in ignore_names
        and not module.file_handle.closed
    ):
        module.file_handle.close()

    module.file_handle = None


def _set_handler(log_file: Union[TextIO, str, None]):
    if isinstance(log_file, str):
        try:
            path = pathlib.Path(log_file).parent.resolve()
            os.makedirs(path, exist_ok=True)
            module.file_handle = open(log_file, "a", encoding="utf-8")
        except Exception as e:
            sys.stderr.write(f"Failed to create log file {log_file} - {e}\n")
    elif isinstance(log_file, io.IOBase):
        module.file_handle = log_file
    elif log_file is None:
        module.file_handle = DEFAULT_AGENT_LOGGER_PATH

    return module.file_handle


def shutdown():
    _close_handler()


def init_structlog(
    log_level_name: str,
    log_file: Union[TextIO, str, None],
    progname: str,
    *,
    # NOTE: We should only enable logger caching if its configuration is finalized. If
    # it's possible for the logging config to change in the future, do not cache it.
    cache_logger: bool = False,
) -> None:
    """
    Initial configuration for structlog. This can still be modified by subsequent calls
    to structlog.configure.
    """
    log_level = _log_level_to_int(log_level_name)

    _close_handler()
    file_handle = _set_handler(log_file)

    structlog.configure(
        processors=[
            structlog.processors.TimeStamper(fmt="iso", key="time"),
            structlog.processors.add_log_level,
            rename_key("event", "msg"),
            structlog.processors.CallsiteParameterAdder(
                [
                    structlog.processors.CallsiteParameter.PROCESS,
                    structlog.processors.CallsiteParameter.FILENAME,
                    structlog.processors.CallsiteParameter.FUNC_NAME,
                    structlog.processors.CallsiteParameter.LINENO,
                    structlog.processors.CallsiteParameter.THREAD,
                    structlog.processors.CallsiteParameter.THREAD_NAME,
                ],
                additional_ignores=[
                    "contrast_vendor.structlog",
                    "contrast.utils.decorators",
                ],
            ),
            rename_key("process", "pid"),
            add_request_id,
            add_asyncio_info,
            add_hostname,
            partial(add_progname, progname=progname),
            structlog.processors.format_exc_info,
            structlog.processors.StackInfoRenderer(
                additional_ignores=[
                    "contrast_vendor.structlog",
                    "contrast.utils.decorators",
                ]
            ),
            structlog.processors.UnicodeDecoder(),
            structlog.processors.JSONRenderer(),
        ],
        logger_factory=structlog.WriteLoggerFactory(file_handle),
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
        cache_logger_on_first_use=cache_logger,
    )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
import os

from threading import Thread
from typing import Optional

from contrast.reporting import Reporter
from contrast.utils.configuration_utils import get_hostname, get_platform
from contrast.reporting.teamserver_messages.server_inventory import ServerInventory
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")


REPORT_RESOURCE_ID_THREAD_NAME = "ContrastReportServerInventory"


def report_server_runtime(
    reporting_client: Reporter, discover_cloud_resource_enabled: bool = True
) -> None:
    """
    Report runtime environment info to TS in a background thread.
    """
    logger.debug("Starting detect runtime environment background thread")
    Thread(
        name=REPORT_RESOURCE_ID_THREAD_NAME,
        daemon=True,
        target=_report_runtime_environment,
        args=(reporting_client, discover_cloud_resource_enabled),
    ).start()


@fail_loudly("Failed to report operating system runtime environment details.")
def _report_runtime_environment(
    reporting_client: Reporter, discover_cloud_resource_enabled
) -> None:
    provider = None
    resource_id = None

    if discover_cloud_resource_enabled:
        provider, resource_id = _detect_cloud_resource_id()

    os_info = get_platform()
    runtime_path = sys.executable
    runtime_version = sys.version.split(" ")[0]
    hostname = get_hostname()
    is_kubernetes = _is_kubernetes()
    is_docker = _is_docker()

    reporting_client.add_message(
        ServerInventory(
            os_info,
            runtime_path,
            runtime_version,
            hostname,
            is_kubernetes=is_kubernetes,
            is_docker=is_docker,
            cloud_provider=provider,
            cloud_resource_id=resource_id,
        )
    )


def _is_kubernetes():
    return "KUBERNETES_SERVICE_HOST" in os.environ


@fail_quietly()
def _is_docker():
    docker_env_file = "/.dockerenv"
    cgroup_file = "/proc/self/cgroup"

    if os.path.isfile(docker_env_file):
        return True

    try:
        with open(cgroup_file) as cgroup:
            return "/docker" in cgroup.read().strip()
    except FileNotFoundError:
        logger.debug("The file %s was not found", cgroup_file)

    return False


def _detect_cloud_resource_id():
    extract_from = {
        "aws": extract_aws_resource_id,
        "azure": extract_azure_resource_id,
        # "gcp": extract_gcp_resource_id, # TODO: PYT-3346
    }

    for provider, extract_id_func in extract_from.items():
        try:
            resource_id = extract_id_func()
        except ResourceIdError as e:
            logger.debug(
                "did not extract resource ID",
                provider=provider,
                exc_info=e,
            )
            continue

        logger.debug(
            "discovered resource ID",
            cloud_provider=provider,
            resource_id=resource_id,
        )
        return provider, resource_id
    return None, None


METADATA_ENDPOINT_ADDRESS = "169.254.169.254"
TOTAL_RETRIES = 5
RETRY_BACKOFF_FACTOR = 1  # wait 0s, 1s, 2s, 4s, 8s, ..., on request retry
NO_PROXIES = {"http": "", "https": ""}
TIMEOUT_SECONDS = 10


class ResourceIdError(Exception):
    pass


def extract_azure_resource_id() -> str:
    """
    Extract the resource ID from the Azure Instance Metadata Service.
    """
    import requests
    from requests.adapters import HTTPAdapter
    from urllib3.util import Retry
    from urllib3.exceptions import MaxRetryError

    # This endpoint comes from https://learn.microsoft.com/en-us/azure/virtual-machines/instance-metadata-service?tabs=linux#route-parameters
    RESOURCE_ID_ENDPOINT = f"http://{METADATA_ENDPOINT_ADDRESS}/metadata/instance/compute/resourceId?api-version=2023-07-01&format=text"

    # https://learn.microsoft.com/en-us/azure/virtual-machines/instance-metadata-service?tabs=linux#errors-and-debugging
    AZURE_RETRY_CODES = {410, 429, 500}

    with requests.Session() as session:
        retries = Retry(
            total=TOTAL_RETRIES,
            backoff_factor=RETRY_BACKOFF_FACTOR,
            status_forcelist=Retry.RETRY_AFTER_STATUS_CODES.union(AZURE_RETRY_CODES),
        )
        session.mount("http://", HTTPAdapter(max_retries=retries))

        try:
            response = session.get(
                RESOURCE_ID_ENDPOINT,
                proxies=NO_PROXIES,
                headers={"Metadata": "true"},
                timeout=TIMEOUT_SECONDS,
            )
            response.raise_for_status()
        except (MaxRetryError, requests.RequestException) as e:
            raise ResourceIdError("Failed to get Azure resource ID") from e
        return response.text


def extract_aws_resource_id() -> str:
    """
    Generate an AWS ARN for the EC2 instance using its Identity Document.

    See:
    https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-identity-documents.html
    """
    logger.debug("Attempting to extract AWS resource ID")

    import requests
    import requests.adapters
    import urllib3.util
    import urllib3.exceptions

    retry_config = urllib3.util.Retry(
        total=TOTAL_RETRIES,
        backoff_factor=RETRY_BACKOFF_FACTOR,
        # force a retry for status codes supporting Retry-After, even if the header is
        # missing
        status_forcelist=urllib3.util.Retry.RETRY_AFTER_STATUS_CODES,
    )

    with requests.Session() as session:
        session.mount(
            "http://", requests.adapters.HTTPAdapter(max_retries=retry_config)
        )
        token = _get_aws_token(session)
        # a token is required for IMDSv2, but if we can't get one, still try IMDSv1 (no
        # token)
        headers = {"X-aws-ec2-metadata-token": token} if token else {}
        try:
            response = session.get(
                f"http://{METADATA_ENDPOINT_ADDRESS}/latest/dynamic/instance-identity/document",
                headers=headers,
                proxies=NO_PROXIES,
                timeout=TIMEOUT_SECONDS,
            )
            response.raise_for_status()
            identity_doc = response.json()
        except (
            urllib3.exceptions.MaxRetryError,
            requests.RequestException,
        ) as e:
            raise ResourceIdError("Failed to get AWS resource ID") from e

    logger.debug("Retrieved AWS identity document", aws_identity_doc=identity_doc)
    try:
        region = identity_doc["region"]
        account_id = identity_doc["accountId"]
        instance_id = identity_doc["instanceId"]
    except KeyError as e:
        raise ResourceIdError(
            "Failed to get AWS resource ID. "
            "Missing a required field in identity document."
        ) from e

    arn = ":".join(
        [
            "arn",
            "aws",
            "ec2",
            region,
            account_id,
            f"instance/{instance_id}",
        ]
    )
    return arn


def _get_aws_token(session) -> Optional[str]:
    logger.debug("Retrieving AWS token")

    import requests
    import urllib3.exceptions

    try:
        token_response = session.put(
            f"http://{METADATA_ENDPOINT_ADDRESS}/latest/api/token",
            headers={
                "X-aws-ec2-metadata-token-ttl-seconds": "300",
            },
            proxies=NO_PROXIES,
            timeout=TIMEOUT_SECONDS,
        )
        token_response.raise_for_status()
        return token_response.text
    except (
        urllib3.exceptions.MaxRetryError,
        requests.RequestException,
    ) as e:
        logger.debug(
            "Unable to retrieve token for AWS IMDSv2 - proceeding without it",
            error=str(e),
        )
    return None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

"""
The sitecustomize.py file is automatically loaded by Python during interpreter
initialization. Our runner script ensures that it is on the PYTHONPATH which is
sufficient to make sure it is loaded.

See https://docs.python.org/3/library/site.html for additional details
"""

### This is the very first line of code we control in the app's python process ###

# We want to minimize the amount of code we run before we register the rewriter,
# because modules imported during that time won't be fully rewritten or patched.
# We attempt to repatch imported modules, but there still could be gaps.

import sys

MIN_SUPPORTED_VERSION = (3, 9)
MAX_SUPPORTED_VERSION = (3, 13)
# We need to check the version_info because certain installation tools bypass
# pip and setuptools, so the version check in setup.py is skipped. For example,
# the agent operator and the universal agent bundle the agent distribution and
# prepend them to PYTHONPATH directly.
if MIN_SUPPORTED_VERSION <= sys.version_info[:2] <= MAX_SUPPORTED_VERSION:
    import contrast_rewriter

    # NOTE: This must be applied prior to importing the agent itself. Do not import any
    # other modules before registering the rewriter.
    contrast_rewriter.register()

    from contrast import loader  # noqa: E402
    from contrast.loader import instrumentation  # noqa: E402

    instrumentation.preinstrument()

    loader.SITECUSTOMIZE_LOADED = True


### Attempt to import the next sitecustomize.py file ###

import os  # noqa: E402

loader_dir = os.path.dirname(__file__)
if loader_dir not in sys.path:
    # loader_dir will always be in sys.path when this sitecustomize is
    # being called by the Python interpreter site startup machinery.
    # If it's missing, we're being imported by some other module at
    # runtime, after the site machinery has already run. In that case,
    # don't attempt to import the next sitecustomize module because
    # that's not what an uninstrumented Python process would do.
    pass
else:
    # Unhook this sitecustomize from the import machinery.
    loader_syspath_index = sys.path.index(loader_dir)
    del sys.path[loader_syspath_index]
    this_sitecustomize = sys.modules["sitecustomize"]
    del sys.modules["sitecustomize"]

    # Attempt to import the next sitecustomize module.
    try:
        import sitecustomize  # noqa: F401
    except ImportError:
        # Only replace sitecustomize if the import fails. Otherwise,
        # an instrumented runner might assume its sitecustomize is in
        # place and raise errors if our sitecustomize were imported
        # instead.
        sys.modules["sitecustomize"] = this_sitecustomize
        sys.path.insert(loader_syspath_index, loader_dir)

    # Only catch the ImportError so that we don't swallow exceptions raised
    # by a buggy third-party sitecustomize implementation.



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.middlewares.route_coverage import common
from contrast_fireball import DiscoveredRoute


def create_routes(app) -> set[DiscoveredRoute]:
    """
    Returns all the routes registered to a Flask or Quart app
    """
    routes = set()

    for rule in list(app.url_map.iter_rules()):
        view_func = app.view_functions[rule.endpoint]
        signature = common.build_signature(rule.endpoint, view_func)
        methods = rule.methods or common.DEFAULT_ROUTE_METHODS
        for method_type in methods:
            routes.add(
                DiscoveredRoute(
                    verb=method_type,
                    url=common.get_normalized_uri(str(rule)),
                    signature=signature,
                    framework=("Quart" if type(app).__name__ == "Quart" else "Flask"),
                )
            )

    return routes



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from dataclasses import replace
from enum import Enum
from typing import Optional

import contrast
from contrast.agent.disable_reaction import DisableReaction
from contrast.agent.request import Request
from contrast.configuration.agent_config import AgentConfig
from contrast.configuration.config_option import DEFAULT_VALUE_SRC
from contrast.reporting.reporting_client import ReportingClient
from contrast import get_canonical_version
import contrast_fireball

from contrast_vendor import structlog as logging, wrapt

logger = logging.getLogger("contrast")


def _handle_errors(return_value=None) -> wrapt.FunctionWrapper:
    """
    A decorator that catches and logs errors that occur while reporting to Contrast.

    Disabling the agent in response to authentication errors or archived applications
    is handled here.

    Errors that indicate a bug in the agent or Fireball are reported to telemetry.

    This decorator should only be used on Client methods, since it expects the
    wrapped function to be a method with AgentConfig stored on the instance.
    """

    @wrapt.function_wrapper
    def wrapper(wrapped, instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except contrast_fireball.Error as e:
            if isinstance(
                e,
                (
                    contrast_fireball.ConfigurationError,
                    contrast_fireball.AuthenticationError,
                    contrast_fireball.AppArchivedError,
                ),
            ):
                # These error messages are user-facing. Log them directly without
                # the stack trace to reduce the noise in the message.
                logger.error(e.message)
            else:
                logger.error(
                    "An error occurred while reporting to Contrast", exc_info=e
                )

            if (
                isinstance(
                    e,
                    (
                        contrast_fireball.Panic,
                        contrast_fireball.ArgumentValidationError,
                        contrast_fireball.UnexpectedError,
                    ),
                )
                and contrast.TELEMETRY is not None
            ):
                contrast.TELEMETRY.report_error(e, wrapped)

            if isinstance(
                e,
                (
                    contrast_fireball.AppArchivedError,
                    contrast_fireball.AuthenticationError,
                ),
            ):
                DisableReaction.run(instance.config)

            return return_value

    return wrapper


class Client(ReportingClient):
    """
    A client for reporting to the Contrast UI using the Fireball library.
    Fireball docs: https://fireball.prod.dotnet.contsec.com/fireball/index.html

    The client will fallback to directly reporting for endpoints that do not
    have Python bindings yet.
    """

    def __init__(self):
        self.config = None
        info = contrast_fireball.get_info()
        super().__init__(instance_id=info["reporting_instance_id"])

    @_handle_errors(return_value=False)
    def initialize_application(self, config: AgentConfig, framework="") -> bool:
        """
        Initialize an application in the Contrast UI.

        This function must be called before any other reporting functions.
        """

        # Store config on the client for disable reaction on AppArchivedError
        self.config = config

        result = contrast_fireball.initialize_application(
            contrast_fireball.InitOptions(
                app_name=config["application.name"],
                app_path=config["application.path"],
                agent_language=contrast_fireball.AgentLanguage.PYTHON,
                agent_version=get_canonical_version(),
                server_host_name=config["server.name"],
                server_path=config["server.path"],
                server_type=config["server.type"] or framework,
                config_paths=None,
                overrides=agent_config_to_plain_dict(config),
            )
        )
        self.app_id = result.data["app_id"]
        config.session_id = result.data["common_config"]["application"].get(
            "session_id", ""
        )

        # This is a workaround to process the startup settings from the UI.
        # Long-term, we'll read the settings from the result, but for now
        # we want to use the well-tested direct response processing behavior.
        return super().initialize_application(config)

    @_handle_errors()
    def new_discovered_routes(self, routes: set[contrast_fireball.DiscoveredRoute]):
        """
        Report discovered routes to the Contrast UI.

        If an exception occurs, no routes are reported.
        """

        contrast_fireball.new_discovered_routes(self.app_id, list(routes))

    @_handle_errors()
    def new_observed_route(self, route: contrast_fireball.ObservedRoute):
        """
        Record an observed route.

        Routes are reported periodically in batches. This endpoint can be called multiple
        times for the same route, but Fireball will only report duplicate routes at a rate
        of once per minute to avoid overloading TeamServer.
        """

        contrast_fireball.new_observed_route(self.app_id, route)

    @_handle_errors()
    def new_findings(
        self,
        findings: list[contrast_fireball.AssessFinding],
        request: Optional[Request],
    ):
        """
        Record Assess findings.

        Findings are reported periodically in batches. Failures are handled for each
        individual finding, so that a failure in one finding does not prevent others
        from being reported.
        """
        fireball_request = request.to_fireball_assess_request() if request else None
        for finding in findings:
            self._new_finding(finding, fireball_request)

    @_handle_errors()
    def _new_finding(
        self,
        finding: contrast_fireball.AssessFinding,
        request: Optional[contrast_fireball.AssessRequest],
    ):
        contrast_fireball.new_finding(self.app_id, replace(finding, request=request))

    @_handle_errors()
    def new_inventory_components(
        self, components: list[contrast_fireball.InventoryComponent]
    ):
        """
        Record Inventory Components.

        Components are reported periodically in batches. Duplicate items between sends
        will be ignored.
        """
        contrast_fireball.new_inventory_components(self.app_id, components)


def agent_config_to_plain_dict(config: AgentConfig):
    """
    Convert all set options in the AgentConfig to a plain dictionary.
    """

    def conv(obj: object):
        if isinstance(obj, Enum):
            return obj.name
        return str(obj)

    json_config = {
        key: conv(v)
        for key, opt in config._config.items()
        if opt.source() != DEFAULT_VALUE_SRC and (v := opt.value()) is not None
    }

    # PROD-1745: Make sure to add the artifact hash to the session metadata.
    # This is implemented in get_session_metadata instead of as the
    # ConfigOption.default for application.session_metadata, because we always
    # want to send the artifact hash even if the user has set other session
    # metadata.
    if "application.session_id" not in json_config:
        json_config["application.session_metadata"] = config.get_session_metadata()

    return json_config



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import time
import threading

from contrast_vendor import importlib_metadata
from contrast_vendor.importlib_metadata import files as importlib_get_files


from contrast.utils.library_reader.utils import get_installed_dist_names
from contrast.utils.library_reader.utils import (
    get_hash,
    CONTRAST_AGENT_DIST,
    PY_SUFFIX,
    SO_SUFFIX,
    is_editable_install,
    normalize_file_name,
)

from contrast.agent import scope
from contrast.api import Library
from contrast.utils.patch_utils import get_loaded_modules
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class LibraryReader:
    def __init__(self, settings, send_ts_message_func):
        self.settings = settings
        self.send_ts_message_func = send_ts_message_func
        self.analysis_thread = None
        self.installed_distribution_keys = []

        # All packages installed in the site-packages/ directory
        self.installed_dists = get_installed_dist_names()

    def start_library_analysis_thread(self, daemon=True):
        self.analysis_thread = threading.Thread(target=self._read_libraries)
        self.analysis_thread.daemon = daemon

        self.analysis_thread.start()

    def join_library_analysis_thread(self):
        if self.analysis_thread:
            self.analysis_thread.join()

    def _read_libraries(self):
        """
        Looks at every library installed in self.installed_dists, then calls search_dist
        on each dist that has metadata associated with the list of files/modules that can be loaded
        :return: None
        """
        with scope.contrast_scope():
            logger.debug("Analyzing libraries...")

            all_dists = read_dists(self.installed_dists)
            reportable_dists = [
                dist
                for dist in all_dists
                if dist.get("file_path", "")
                and dist.get("file_path", "") != CONTRAST_AGENT_DIST
            ]

            self._send_analysis_results_appstart(reportable_dists)
            self._send_analysis_results_files_loaded(reportable_dists)

    def _send_analysis_results_appstart(self, reportable_dists):
        """
        Send library discovery in ApplicationUpdate message.
        :param reportable_dists: A list of dictionaries containing information about
            each distribution.
        """
        if not reportable_dists:
            return

        libraries = [Library(dist) for dist in reportable_dists]

        logger.debug(
            "Discovered libraries: %s",
            [f"{lib.file_path} - {lib.version}" for lib in libraries],
        )

        logger.debug("Sending ApplicationUpdate message with library analysis results")

        # Import here to prevent circular import
        from contrast.reporting import teamserver_messages

        app_update_msg = teamserver_messages.ApplicationUpdate(libraries)
        self.send_ts_message_func(app_update_msg)

    def _send_analysis_results_files_loaded(self, reportable_dists):
        if not reportable_dists:
            return

        from contrast.reporting import teamserver_messages

        msg = teamserver_messages.LibraryUsage(reportable_dists)

        logger.debug("Sending list of modules loaded to Contrast")

        self.send_ts_message_func(msg)


def read_dists(installed_dists):
    results = []

    for dist in installed_dists:
        analysis_result = search_dist(dist)

        if analysis_result:
            results.append(analysis_result)

    return results


def search_dist(dist):
    """
    Searches directories related to dist, gathering relevant statistics and metadata.
    Then, assuming library was loaded, appends that metadata to the self.libraries.

    Created package is added to the output for the process

    :param dist: Name of the python distribution we are analyzing
    :return: A dictionary containing contrast-relevant information about the current distribution
    """
    excluded_files = ("setup.py",)
    metadata = importlib_metadata.metadata(dist)
    version = metadata.get("Version", "")
    url = metadata.get("Home-page", "")
    all_files = importlib_get_files(dist)
    path = None

    if all_files is None:
        # In the case where the metadata files listing files (RECORD, SOURCES.txt etc..) are missing, importlib_get_files() will return None.
        return None

    if is_editable_install(dist, version, all_files):
        # We decided to omit this case since whatever the package is, its probably under active development
        # and still not a library at that point in time
        return None

    # Omit metadata files from total module count
    all_modules = {
        path
        for f in all_files
        if f.name.endswith((PY_SUFFIX, SO_SUFFIX))
        and f.name not in excluded_files
        and (path := os.path.realpath(str(f.locate())))
    }

    file_count = len(all_modules)
    library_hash = get_hash(dist, version)

    used_files = search_modules(all_modules)

    result = create_package(version, file_count, dist, url, library_hash)
    result["used_files"] = used_files

    return result


def create_package(version, class_count, name, url, sha):
    """
    Generate a json-serializable representation of a package for teamserver

    :param version: version of package
    :param class_count: count of files
    :param name: name of the library
    :param url: homepage of package
    :param sha: sha1 hash of the name(space)version
    :return: dict of package
    """
    current_time = int(time.time() * 1000)

    return {
        "version": version,
        "class_count": class_count,
        "file_path": name,
        "url": url,
        "hash_code": sha,
        "external_ms": current_time,
        "internal_ms": current_time,
    }


def search_modules(all_modules):
    """
    Searches every module in a distribution to see if it is loaded in the python interpreter.

    :param all_modules: modules in a specific package
    :return: The set of modules for a specific dist loaded in the python interpreter
    """
    # __file__ (Path to the loaded file) __file__ is optional. The import system may opt to leave
    # __file__ unset if it has no semantic meaning (e.g. a module loaded from a database).
    # Using os.path.realpath because in some environments there is a symlink to the directory containg 3rd party packages
    loaded_modules = {
        normalize_file_name(os.path.realpath(mod.__file__))
        for mod in get_loaded_modules().values()
        if hasattr(mod, "__file__")
        and mod.__file__
        and mod.__file__ in all_modules
        and normalize_file_name(os.path.realpath(mod.__file__)) is not None
    }

    return loaded_modules



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from functools import cached_property
from typing import Optional
import contrast
from contrast.agent import scope
from contrast.agent.settings import Settings
from contrast.api.attack import Attack, ProtectResponse
from contrast.api.user_input import UserInput, InputType
from contrast.api.sample import Sample
from contrast.agent.protect.rule.mode import Mode
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast.utils.stack_trace_utils import build_and_clean_stack
from contrast_vendor import structlog as logging
from contrast.utils.loggers.logger import security_log_attack
from contrast.utils.string_utils import ensure_string

logger = logging.getLogger("contrast")


BLOCKING_RULES = frozenset([Mode.BLOCK, Mode.BLOCK_AT_PERIMETER])
PREFILTER_RULES = frozenset([Mode.BLOCK_AT_PERIMETER])
POSTFILTER_RULES = frozenset([Mode.BLOCK, Mode.MONITOR])

PREFILTER = "prefilter"
INFILTER = "infilter"
POSTFILTER = "postfilter"


class BaseRule:
    """
    Base rule object that all protection rules will inherit
    """

    RULE_NAME = "base-rule"

    def __init__(self):
        self.settings = Settings()
        self.settings.protect_rules[self.name] = self
        self.probe_analysis_enabled = self.settings.config.get(
            "protect.probe_analysis.enable"
        )

    @property
    def name(self):
        return self.RULE_NAME

    @property
    def mode(self):
        """
        Return the mode for this rule based.

        Order of precedence:
        1. Config (contract_security.yaml)
        2. Settings from TS
        3. Default mode
        """
        return self.settings.config.get(self.config_rule_path_mode)

    @cached_property
    def config_rule_path_mode(self):
        return f"protect.rules.{self.name}.mode"

    def is_prefilter(self):
        """
        Checks if a rules mode is for prefilter
        """
        return self.enabled and self.mode in PREFILTER_RULES

    def is_postfilter(self):
        """
        Checks if a rules mode is for postfilter
        """
        return self.enabled and self.mode in POSTFILTER_RULES

    def is_blocked(self):
        """
        Checks if a rules mode is for blocking
        """
        return self.enabled and self.mode in BLOCKING_RULES

    @property
    def enabled(self):
        """
        A rule is enabled only if all 3 conditions are met:
        1. protect is enabled in both config and server settings
        2. rule is not in disabled rules list
        3. rule mode is not OFF
        4. an exclusion wasn't applied from Teamserver
        """
        if not self.settings.is_protect_enabled():
            return False

        disabled_rules = self.settings.config.get("protect.rules.disabled_rules")
        if disabled_rules and self.name in disabled_rules:
            return False

        req_ctx = contrast.CS__CONTEXT_TRACKER.current()
        if req_ctx is not None and req_ctx.excluded_protect_rules:
            return self.name not in req_ctx.excluded_protect_rules

        return self.mode != Mode.OFF

    def should_block(self, attack):
        return attack and attack.response == ProtectResponse.BLOCKED

    def excluded(self, exclusions):
        """
        Check if rule is being excluded from evaluation
        :param exclusions:
        :return: True if excluded, else False
        """

        if not exclusions or len(exclusions) == 0:
            return False

        logger.debug("Checking %s exclusion(s) in %s", len(exclusions), self.name)
        return any(ex.match_protect_rule(self.name) for ex in exclusions)

    def prefilter(self):
        """
        Scans the input analysis for the rule and looks for matched attack signatures

        Will throw a SecurityException if a response needs to be blocked
        """
        logger.debug("PROTECT: Prefilter for %s", self.name)

        attack = self.find_attack(analysis_stage=PREFILTER)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if attack.response == ProtectResponse.BLOCKED_AT_PERIMETER:
            raise contrast.SecurityException(rule_name=self.name)

    def _infilter(self, match_string, **kwargs):
        """
        Scans the input analysis for the rule and looks for matched attack signatures. The call to this method may be
        rule specific and include additional context in a args list.
        """
        if self.mode == Mode.OFF:
            return

        logger.debug("PROTECT: Infilter for %s", self.name)

        attack = self.find_attack(match_string, analysis_stage=INFILTER, **kwargs)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if self.should_block(attack):
            raise contrast.SecurityException(rule_name=self.name)

    @fail_loudly("Failed to run protect rule")
    def protect(self, patch_policy, user_input, args, kwargs):
        if not self.enabled:
            return

        if self.skip_protect_analysis(user_input, args, kwargs):
            return

        with scope.contrast_scope():
            user_input = self.convert_input(user_input)
            if not user_input:
                return

            self.log_safely(patch_policy.method_name, user_input)

            self._infilter(user_input, **self.infilter_kwargs(user_input, patch_policy))

    def infilter_kwargs(self, user_input, patch_policy):
        return {}

    def skip_protect_analysis(self, user_input, args, kwargs):
        """
        We only want to run protect on user input that is of a type supported
        by the rule.

        Most rules use this implementation, but some override this depending on
        expected user input types.

        :return: Bool if to skip running protect infilter
        """
        if not user_input:
            return True

        if isinstance(user_input, (str, bytes)):
            return False

        logger.debug(
            "WARNING: unknown input type %s for rule %s", type(user_input), self.name
        )

        return True

    def convert_input(self, user_input):
        return ensure_string(user_input)

    def postfilter(self):
        """
        Scans the input analysis for the rule and looks for matched attack signatures

        Appends attacker to the context if a positive evaluation is found
        """
        if self.mode == Mode.OFF or not self.probe_analysis_enabled:
            return

        logger.debug("PROTECT: Postfilter", rule=self.name)

        attack = self.find_attack(analysis_stage=POSTFILTER)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if self.should_block(attack):
            raise contrast.SecurityException(rule_name=self.name)

    def find_attack(self, candidate_string=None, analysis_stage=None, **kwargs):
        """
        Finds the attacker in the original string if present
        """
        if candidate_string is not None:
            logger.debug("Checking for %s in %s", self.name, candidate_string)

        evaluations_for_rule = self.evaluations_for_rule()

        attack = None
        for evaluation in evaluations_for_rule:
            if analysis_stage == POSTFILTER and evaluation.attack_count > 0:
                continue
            if candidate_string:
                if candidate_string.find(evaluation.value) == -1:
                    continue

                attack = self.build_attack_with_match(
                    candidate_string, evaluation, attack, **kwargs
                )
            else:
                attack = self.build_attack_without_match(evaluation, attack, **kwargs)

        return attack

    def build_attack_with_match(
        self,
        candidate_string,
        evaluation=None,
        attack: Optional[Attack] = None,
        **kwargs,
    ):
        attack = self.build_or_append_attack(
            evaluation, attack, candidate_string, **kwargs
        )

        if evaluation:
            evaluation.attack_count += 1

        attack.response = self.response_from_mode(self.mode)
        security_log_attack(attack, evaluation)
        return attack

    def build_attack_without_match(
        self, evaluation=None, attack: Optional[Attack] = None, **kwargs
    ):
        if evaluation and evaluation.score < 10:
            return None
        if self.mode == Mode.BLOCK_AT_PERIMETER:
            attack = self.build_or_append_attack(evaluation, attack, **kwargs)

            attack.response = self.response_from_mode(self.mode)
            security_log_attack(attack, evaluation)
        elif evaluation is None or evaluation.attack_count == 0:
            attack = self.build_or_append_attack(evaluation, attack, **kwargs)
            attack.response = ProtectResponse.PROBED
            security_log_attack(attack, evaluation)

        return attack

    def build_or_append_attack(
        self,
        evaluation,
        attack: Optional[Attack] = None,
        candidate_string=None,
        **kwargs,
    ):
        if attack is None:
            attack = self.build_base_attack()

        attack.add_sample(self.build_sample(evaluation, candidate_string, **kwargs))

        return attack

    def build_base_attack(self):
        return Attack(self.name)

    def build_sample(self, evaluation, candidate_string, **kwargs):
        return self.build_base_sample(evaluation)

    def build_user_input(self, evaluation):
        return UserInput(
            input_type=evaluation.input_type,
            key=evaluation.key,
            value=evaluation.value,
            path=evaluation.path,
            matcher_ids=[idx for idx in evaluation.ids],
        )

    def build_base_sample(self, evaluation, prebuilt_stack=None):
        sample = Sample()

        if evaluation:
            sample.set_user_input(self.build_user_input(evaluation))
        else:
            sample.set_user_input(
                UserInput(input_type=InputType.UNKNOWN, key="", value="")
            )

        stack = prebuilt_stack if prebuilt_stack else build_and_clean_stack()
        sample.set_stack(stack)

        return sample

    def _append_to_context(self, attack):
        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            # do not remove; this case is not yet well-understood
            logger.debug("WARNING: failed to get request context in _append_to_context")
            return

        context.attacks.append(attack)

    _RESPONSE_MAP = {
        Mode.MONITOR: ProtectResponse.MONITORED,
        Mode.BLOCK: ProtectResponse.BLOCKED,
        Mode.BLOCK_AT_PERIMETER: ProtectResponse.BLOCKED_AT_PERIMETER,
        Mode.OFF: ProtectResponse.NO_ACTION,
    }

    def response_from_mode(self, mode):
        return self._RESPONSE_MAP.get(mode)

    def evaluations_for_rule(self, context=None):
        if context is None:
            context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            # do not remove; this case is not yet well-understood
            logger.debug(
                "WARNING: failed to get request context in evaluations_for_rule"
            )
            return []

        return [
            evaluation
            for evaluation in context.user_input_analysis
            if evaluation.rule_id == self.RULE_NAME
        ]

    @fail_quietly("Failed to log user input for protect rule")
    def log_safely(self, method_name, user_input):
        """
        Attempt to log user supplied input but do not fail if unable to do so.
        """
        logger.debug(
            "Applying %s rule method %s with user input %s",
            self.name,
            method_name,
            ensure_string(user_input, errors="replace"),
        )



from argparse import ArgumentParser
from collections import OrderedDict
from datetime import datetime

from . import hookspy
from .body import create_hook
from .signatures import get_method_signature, fastcall_typedef, fastcall_kwargs_typedef


STREAM_TYPES = ["bytesio", "stringio", "iobase"]
TYPES = STREAM_TYPES

stream_methods = [
    "read",
    "readline",
    "readlines",
    "read1",
]


stream_hook_template = """    ADD_STREAM_HOOK(StreamType, {1}, {2});"""
stream_reverse_hook_template = """    REVERSE_STREAM_HOOK(StreamType, {1}, {2});"""


stream_patch_template = """
int patch_{}_methods(PyTypeObject *StreamType) {{

{}

    /* Indicates success */
    return 0;
}}
"""

stream_reverse_template = """
void reverse_{}_methods(PyTypeObject *StreamType) {{

{}

}}
"""

YEAR = datetime.now().year

C_HEADER = f"""/*
* Copyright © {YEAR} Contrast Security, Inc.
* See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
*/
"""

source_file_header = f"""{C_HEADER}/* THIS FILE WAS AUTOMATICALLY GENERATED BY HOOKSPY */
/* Python requires its own header to always be included first */
#define PY_SSIZE_T_CLEAN
#include <Python.h>

#include <contrast/assess/patches.h>
#include <contrast/assess/propagate.h>
#include <contrast/assess/scope.h>
#include <contrast/assess/utils.h>

{fastcall_typedef};
{fastcall_kwargs_typedef};
"""


pytype_map = {
    "bytesio": hookspy.find_bytesio_hook,
    "stringio": hookspy.find_stringio_hook,
    "iobase": hookspy.find_iobase_hook,
}


def transform_name(name):
    # This is kind of annoying, but the name of this method in the C-level
    # method table has an underscore.
    return name.replace("formatmap", "format_map")


def find_hooks(spyfunc, methods):
    hook_methods = OrderedDict()
    for name in methods:
        result = spyfunc(transform_name(name))
        if result is None:
            continue
        hook_methods[name] = result
    return hook_methods


def raise_unsupported_type(typename):
    raise ValueError(f"Unsupported type: {typename}")


def autogen(typename):
    spyfunc = pytype_map[typename]
    stream_type = typename in STREAM_TYPES
    methods = stream_methods if stream_type else []
    hook_template = (
        stream_hook_template if stream_type else raise_unsupported_type(typename)
    )
    patch_template = (
        stream_patch_template if stream_type else raise_unsupported_type(typename)
    )

    macros = []
    declarations = []
    hook_funcs = []
    hook_points = []
    reverse_points = []
    hook_methods = find_hooks(spyfunc, methods)

    for name, (offset, argtype) in hook_methods.items():
        signature = get_method_signature(argtype)

        hook_definition, hook_name = create_hook(typename, name, signature, stream_type)
        declarations.append(f"{signature} {hook_name}_orig;")
        hook_funcs.append(hook_definition)

        macro_name = f"{typename}_{name}_OFFSET".upper()
        macros.append(f"#define {macro_name} {offset}")

        hook_points.append(hook_template.format("", hook_name, macro_name))
        reverse_points.append(
            stream_reverse_hook_template.format("", hook_name, macro_name)
        )

    print(source_file_header)
    for m in macros:
        print(m)
    print("\n")
    for d in declarations:
        print(d)
    print("\n")
    for f in hook_funcs:
        print(f)
    print("\n")
    print(patch_template.format(typename, "\n".join(hook_points)))

    if stream_type:
        print(stream_reverse_template.format(typename, "\n".join(reverse_points)))


def main():
    p = ArgumentParser("Automatically generate Python type hooks")
    p.add_argument("type", type=str, help="Python type", choices=TYPES)
    args = p.parse_args()

    autogen(args.type)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.middlewares.route_coverage.common import (
    DEFAULT_ROUTE_METHODS,
    get_normalized_uri,
)
from contrast.agent.middlewares.route_coverage.common import build_signature
from contrast_fireball import DiscoveredRoute

DEFAULT_ROUTE_METHODS = DEFAULT_ROUTE_METHODS + ("HEAD",)


def create_starlette_routes(starlette_router) -> set[DiscoveredRoute]:
    """
    Returns all the routes registered to a Starlette router as a dict.
    :param starlette_app: Starlette router instance (starlette.routing.Router)
    :return: dict {route_id:  api.Route}
    """
    from starlette.routing import Mount, Route

    routes = set()

    for app_route in starlette_router.routes:
        if isinstance(app_route, Mount):
            mnt_routes = create_starlette_routes(app_route)
            routes.update(mnt_routes)
            continue

        elif isinstance(app_route, Route):
            view_func = app_route.endpoint

            signature = build_signature(app_route.name, view_func)
            methods = app_route.methods or DEFAULT_ROUTE_METHODS

            for method_type in methods:
                routes.add(
                    DiscoveredRoute(
                        verb=method_type,
                        url=get_normalized_uri(str(app_route.name)),
                        signature=signature,
                        framework="Starlette",
                    )
                )
        else:
            # This is a catch-all for other BaseRoute types which we
            # don't support, such as WebSocketRoute
            continue

    return routes



from .signatures import fastcall_name


hook_macro_by_signature = {
    "unaryfunc": "HOOK_{}UNARYFUNC",
    "binaryfunc": "HOOK_{}BINARYFUNC",
    fastcall_name: "HOOK_{}FASTCALL",
}


def build_stream_hook_macro(name, hook_name, signature):
    hook_definition = hook_macro_by_signature[signature]
    hook_definition = hook_definition.format("STREAM_")
    hook_definition += '({}, "{}");'
    return hook_definition.format(hook_name, name)


def create_hook(strtype, name, signature, stream_type):
    hook_name = f"{strtype}_{name}"
    if stream_type:
        hook_definition = build_stream_hook_macro(name, hook_name, signature)
    else:
        raise ValueError(f"Unknown hook {hook_name}")

    return hook_definition, hook_name



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import functools
import os
import re
import hashlib

from contrast_vendor.importlib_metadata import packages_distributions
from contrast_vendor.importlib_metadata import files as importlib_get_files

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

CONTRAST_AGENT_DIST = "contrast-agent"

# Both of these metadata files contain a file list of what is installed under the top level dirs
RECORD = "RECORD"
SOURCES = "SOURCES.txt"

NAMESPACE_PACKAGE = "namespace_packages.txt"
TOP_LEVEL_TXT = "top_level.txt"

PY_SUFFIX = ".py"
SO_SUFFIX = ".so"

SITE_PACKAGES_DIR = f"{os.sep}site-packages{os.sep}"
DIST_PACKAGES_DIR = f"{os.sep}dist-packages{os.sep}"


def _load_installed_dist_top_level_imports(
    cache: dict[str, set[str]],
) -> dict[str, set[str]]:
    """
    Returns a dictionary containing a mapping of the package name to each of its top level importable packages by name.
    """
    if cache is None:
        cache = {}

    for top_level_import, dist_names in packages_distributions().items():
        dist_names = set(dist_names)

        for dist in dist_names:
            if cache.get(dist, None) is None:
                cache[dist] = {
                    top_level_import,
                }
            else:
                cache[dist].add(top_level_import)

    return cache


@functools.lru_cache
def get_installed_dist_top_level_imports():
    return _load_installed_dist_top_level_imports({})


@functools.lru_cache
def get_installed_dist_names():
    return list(get_installed_dist_top_level_imports().keys())


def is_editable_install(dist, version, all_files):
    editable_install_metadata_fname = f"__editable__{dist.lower()}-{version}.pth"

    return editable_install_metadata_fname in all_files


def normalize_file_name(file_path):
    """
    This function converts a file ending in .pyc to .py. The reason for this
    is due to how screener is configured to verify a file was reported (only supports
    exact match not a regex)
    @param file_path: full path to a python file ending in .pyc or .py
    @return: file_path ending in .py
    """
    file_to_report = file_path.rpartition(SITE_PACKAGES_DIR)

    if not file_to_report[1]:
        file_to_report = file_path.rpartition(DIST_PACKAGES_DIR)
        if not file_to_report[1]:
            return None

    normalized_file_name = file_to_report[2]
    if normalized_file_name.endswith(".pyc"):
        normalized_file_name = normalized_file_name[: len(normalized_file_name) - 1]

    return normalized_file_name


def get_top_level_directories_namespace_pkg(dist: str, namespace: str) -> set:
    """
    @param dist: Distribution name
    @param namespace: The name of the namespace to search
    @return: The top level importable packages/modules under the namespace
    """
    top_level_dirs = set()
    ignore_dirs = ("__pycache__",)
    excluded_files = ("setup.py",)
    all_files = importlib_get_files(dist)

    if not dist:
        return top_level_dirs

    if all_files is None:
        # In the case where the metadata files listing files (RECORD, SOURCES.txt etc..) are missing, importlib_get_files() will return None.
        return top_level_dirs

    # Go through each python file and get the directory name after namespace
    for f in all_files:
        current_file = os.path.realpath(str(f.locate()))
        if current_file:
            current_file = normalize_file_name(current_file)

        if (
            current_file
            and f.name.endswith((PY_SUFFIX, SO_SUFFIX))
            and f.name not in excluded_files
            and current_file.startswith(namespace + os.sep)
        ):
            dirs = current_file.split(os.sep)
            if len(dirs) > 1 and dirs[1] not in ignore_dirs:
                top_level_dirs.add(dirs[1])

    return top_level_dirs


def get_file_from_module(module):
    if hasattr(module, "__file__") and module.__file__:
        return os.path.realpath(module.__file__)

    return None


def normalize_dist_name(name):
    # Normalize the name based on the spec https://packaging.python.org/en/latest/specifications/name-normalization/
    return re.sub(r"[-_.]+", "-", name).lower()


def get_hash(name, version):
    """
    DO NOT ALTER OR REMOVE

    This must match the calculation made by the artifact dependency management database.
    """
    name = normalize_dist_name(name)

    to_hash = name + " " + version

    return hashlib.sha1(to_hash.encode("utf-8")).hexdigest()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections.abc import Generator, Mapping
from typing import Optional, Protocol


from contrast.agent.request import Request
from contrast.configuration.agent_config import AgentConfig
from contrast.reporting.teamserver_messages.base_ts_message import BaseTsMessage
from contrast_fireball import (
    AssessFinding,
    DiscoveredRoute,
    InventoryComponent,
    ObservedRoute,
)


class Reporter(Protocol):
    def initialize_application(self, config: AgentConfig, framework="") -> bool: ...

    def new_discovered_routes(self, routes: set[DiscoveredRoute]): ...

    def new_observed_route(self, route: ObservedRoute): ...

    # new_findings is a batching method, but the Fireball client
    # will accept a single finding at a time and batch them internally.
    # When Fireball is the primary reporting client, we should consider
    # moving findings to a fire-and-forget model instead of batching.
    def new_findings(
        self, findings: list[AssessFinding], request: Optional[Request]
    ): ...

    def new_inventory_components(self, components: list[InventoryComponent]): ...

    def observability_trace_context(self) -> Generator[None, None, None]: ...

    def new_observability_span(self, span: dict) -> None: ...

    def update_observability_metric(self, metric: dict) -> None: ...

    # Legacy methods from direct reporting
    def add_message(self, msg: BaseTsMessage): ...

    def send_message(self, msg: BaseTsMessage): ...

    def retry_message(self, msg: BaseTsMessage): ...


def get_reporting_client(config: Mapping) -> Reporter:
    client_type = config.get("api.reporting_client")
    if client_type == "fireball":
        from contrast.reporting.fireball import Client
    elif client_type == "direct":
        from contrast.reporting.reporting_client import ReportingClient as Client
    else:
        raise ValueError(f"Invalid reporting client: {client_type}")

    return Client()



from sys import version_info

from . import hookspy


py_version = version_info[:2]


fastcall_name = "fastcall_func"
fastcall_typedef = (
    f"typedef PyObject *(*{fastcall_name})(PyObject *, PyObject *const *, Py_ssize_t)"
)

fastcall_kwargs_name = "fastcall_kwargs_func"
fastcall_kwargs_typedef = f"typedef PyObject *(*{fastcall_kwargs_name})(PyObject *, PyObject *const *, Py_ssize_t, PyObject *)"


unaryfunc_signature = "PyObject *{}(PyObject *self)"
binaryfunc_signature = "PyObject *{}(PyObject *self, PyObject *args)"
ternaryfunc_signature = "PyObject *{}(PyObject *self, PyObject *args, PyObject *kwargs)"
fastcall_signature = (
    "PyObject *{}(PyObject *self, PyObject *const *args, Py_ssize_t nargs)"
)
fastcall_kwargs_signature = "PyObject *{}(PyObject *self, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames)"

newfunc_signature = "PyObject *{}(PyTypeObject *type, PyObject *args, PyObject *kwds)"
initproc_signature = "int {}(PyObject *self, PyObject *args, PyObject *kwds)"

unaryfunc_call = "(self)"
binaryfunc_call = "(self, args)"
ternaryfunc_call = "(self, args, kwargs)"
fastcall_call = "(self, args, nargs)"
fastcall_kwargs_call = "(self, args, nargs, kwargs)"
newfunc_call = "(type, args, kwds)"
initproc_call = "(self, args, kwds)"


propagate_args_by_name = {
    "unaryfunc": "NULL, NULL",
    "binaryfunc": "args, NULL",
    "ternaryfunc": "args, kwargs",
}


method_signatures_map = {
    hookspy.METH_O: "binaryfunc",
    hookspy.METH_NOARGS: "unaryfunc",
    hookspy.METH_VARARGS: "binaryfunc",
    hookspy.METH_KEYWORDS: "ternaryfunc",
    hookspy.METH_FASTCALL: fastcall_name,
    hookspy.METH_VARARGS | hookspy.METH_KEYWORDS: "ternaryfunc",
    hookspy.METH_FASTCALL | hookspy.METH_KEYWORDS: fastcall_kwargs_name,
}


def get_method_signature(argtype):
    return method_signatures_map[argtype]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Optional

from contrast_fireball import (
    AssessFinding,
    AssessRoute,
    AssessRouteObservation,
    ObservedRoute,
)
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class RoutesMixin:
    """
    Class for route coverage work.
    Route coverage is assess only.
    """

    def handle_routes(self, context, request) -> None:
        """
        Method that should run for all middlewares at the end of request processing.
        """
        if context and request:
            self.update_route_information(context, request)
            self.append_route_to_findings(context.observed_route, context.findings)

    def update_route_information(self, context, request) -> None:
        """
        Store the information for this request's route on RequestContext.
        """
        context.observed_route = ObservedRoute(
            signature=self._get_signature(context, request) or "",
            url=context.request.get_normalized_uri(),
            verb=request.method,
            sources=context.observed_route.sources,
        )

        logger.debug(
            "stored observed route on context", observed_route=context.observed_route
        )

    def _get_signature(self, context, request) -> Optional[str]:
        """
        There are a few different strategies we might use to obtain a route signature.
        In order of preference:

        1.  Obtain a signature at some point during the request; this string is set on
            `context.view_func_str`. For supported frameworks, this is accomplished with
            framework-specific patches. One exception to this is aiohttp, where it
            happens directly in the middleware. It is essential that for a particular
            route this signature exactly matches the one found during route discovery.
            In the future, all frameworks will take this approach.

        2.  If context.view_func_str isn't set, we didn't get any framework-specific
            information about the view function. In this case we use the request's
            normalized URI as the signature. We expect to hit this case in pure WSGI /
            ASGI.

        Framework-specific route observation code should set either view_func_str, which
        takes priority, or view_func. It doesn't make sense to set both, since only
        view_func_str would be used in that case.
        """
        if context.view_func_str:
            return context.view_func_str

        logger.debug(
            "Did not find a view function signature for the current request. "
            "Falling back on normalized URI."
        )
        return context.request.get_normalized_uri()

    def append_route_to_findings(
        self, observed_route: ObservedRoute, findings: list[AssessFinding]
    ):
        """
        Append the observed route to any existing findings. We can't necessarily do this
        at finding creation time, because we might not have route info yet.
        """
        if not findings:
            logger.debug("No findings to append route to")
            return

        for finding in findings:
            if not finding.routes:
                logger.debug(
                    "Appending route %s:%s to %s",
                    observed_route.verb,
                    observed_route.url,
                    finding.rule_id,
                )
                finding.routes.append(
                    AssessRoute(
                        count=1,
                        signature=observed_route.signature,
                        observations=[
                            AssessRouteObservation(
                                url=observed_route.url, verb=observed_route.verb or ""
                            )
                        ],
                    )
                )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.middlewares.route_coverage.common import (
    build_signature,
    DEFAULT_ROUTE_METHODS,
    get_normalized_uri,
)
from contrast_fireball import DiscoveredRoute

DEFAULT_ROUTE_METHODS = DEFAULT_ROUTE_METHODS + ("PUT", "PATCH", "DELETE")


def create_bottle_routes(app) -> set[DiscoveredRoute]:
    """
    Returns all the routes registered to a Bottle app.
    """
    return {
        DiscoveredRoute(
            verb=method_type,
            url=get_normalized_uri(str(route.rule)),
            signature=build_signature(route.rule, route.callback),
            framework="Bottle",
        )
        for route in app.routes
        for method_type in DEFAULT_ROUTE_METHODS
    }
